/**
 *
 * Copyright (C) 2013 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/
// Blob
// Changelog 05.01.24

// Standard include
#include "blob.h"

/*!
\class BlobBlend blob.h
\brief This class implements a blending node in the blobby scene graph.
\ingroup BlobNodeGroup 
*/

/*! 
\brief Create a blending node given a set of children nodes.
*/
BlobBlend::BlobBlend(BlobNode* a,BlobNode* b)
{
  elements[0]=a; 
  elements[1]=b;

  box=Box(elements[0]->GetBox(),elements[1]->GetBox());
}

/*!
\brief Destroys a blending node.
*/
BlobBlend::~BlobBlend()
{
  for (int i=0;i<2;i++)
  {
    delete elements[i];
  }
}

/*! 
\brief Computes the field value at a given point in space.
*/ 
double BlobBlend::Intensity(const Vector& p)
{  
  // Box 
  if (!box.Inside(p)) return 0.0;

  return elements[0]->Intensity(p)+elements[1]->Intensity(p);
}

Vector BlobBlend::Color(const Vector& p){  
  if(elements[0]->Intensity(p)>elements[1]->Intensity(p))
    return elements[0]->Color(p);
  return elements[1]->Color(p);
}




/*! 
\brief Create a blending node given a set of children nodes.
*/
BlobUnion::BlobUnion(BlobNode* a,BlobNode* b)
{
  elements[0]=a; 
  elements[1]=b;

  box=Box(elements[0]->GetBox(),elements[1]->GetBox());
}

/*!
\brief Destroys a blending node.
*/
BlobUnion::~BlobUnion()
{
  for (int i=0;i<2;i++)
  {
    delete elements[i];
  }
}

/*! 
\brief Computes the field value at a given point in space.
*/ 
double BlobUnion::Intensity(const Vector& p)
{  
  // Box 
  if (!box.Inside(p)) return 0.0;

  return std::max (  elements[0]->Intensity(p) , elements[1]->Intensity(p));

}

Vector BlobUnion::Color(const Vector& p){  
if(elements[0]->Intensity(p)>elements[1]->Intensity(p))
    return elements[0]->Color(p);
  return elements[1]->Color(p);
}



/*! 
\brief Create a blending node given a set of children nodes.
*/
BlobIntersect::BlobIntersect(BlobNode* a,BlobNode* b)
{
  elements[0]=a; 
  elements[1]=b;

  box=Box(elements[0]->GetBox(),elements[1]->GetBox());
}

/*!
\brief Destroys a blending node.
*/
BlobIntersect::~BlobIntersect()
{
  for (int i=0;i<2;i++)
  {
    delete elements[i];
  }
}
/*! 
\brief Computes the field value at a given point in space.
*/ 
double BlobIntersect::Intensity(const Vector& p)
{  
  // Box 
  if (!box.Inside(p)) return 0.0;

  return std::min (  elements[0]->Intensity(p) , elements[1]->Intensity(p));
}

Vector BlobIntersect::Color(const Vector& p){  
if(elements[0]->Intensity(p)>elements[1]->Intensity(p))
    return elements[0]->Color(p);
  return elements[1]->Color(p);
}


/*! 
\brief Create a blending node given a set of children nodes.
*/
BlobSub::BlobSub(BlobNode* a,BlobNode* b)
{
  elements[0]=a; 
  elements[1]=b;

  box=Box(elements[0]->GetBox(),elements[1]->GetBox());
}

/*!
\brief Destroys a blending node.
*/
BlobSub::~BlobSub()
{
  for (int i=0;i<2;i++)
  {
    delete elements[i];
  }
}
/*! 
\brief Computes the field value at a given point in space.
*/ 
double BlobSub::Intensity(const Vector& p){  
  // Box 
  if (!box.Inside(p)) return 0.0;
  

  return elements[0]->Intensity(p)-elements[1]->Intensity(p);
  
}

Vector BlobSub::Color(const Vector& p){  
  if(elements[1]->Intensity(p)>=1e-4)
    return elements[1]->Color(p);
  else return elements[0]->Color(p);
}


/*! 
\brief Create a blending node given a set of children nodes.
*/
BlobXor::BlobXor(BlobNode* a,BlobNode* b){
  elements[0]=a; 
  elements[1]=b;
  box=Box(elements[0]->GetBox(),elements[1]->GetBox());
}

/*!
\brief Destroys a blending node.
*/
BlobXor::~BlobXor()
{
  for (int i=0;i<2;i++)
  {
    delete elements[i];
  }
}
/*! 
\brief Computes the field value at a given point in space.
*/ 
double BlobXor::Intensity(const Vector& p){  
  // Box 
  if (!box.Inside(p)) return 0.0;
  
  if(elements[0]->Intensity(p) and elements[1]->Intensity(p))
    return elements[0]->Intensity(p)-elements[1]->Intensity(p);
  
  if(elements[0]->Intensity(p))
    return elements[0]->Intensity(p);

  if(elements[1]->Intensity(p))
    return elements[1]->Intensity(p);

  return 0.0;
}

Vector BlobXor::Color(const Vector& p){  
if(elements[0]->Intensity(p)>elements[1]->Intensity(p))
    return elements[0]->Color(p);
  return elements[1]->Color(p);
}
