#include "obbox.h"

// --------------------------
//
// Oriented Bounding Box Class
//
// --------------------------

//
// Check if a point is in this bounding box
//
bool CBBox::IsPointInBox(const Vector &InP)
{
  // Rotate the point into the box's coordinates
  Vector P = m_M.InvertSimple() * InP;
	
  // Now just use an axis-aligned check
  if ( fabs(P[0]) < m_Extent[0] && fabs(P[1]) < m_Extent[1] && fabs(P[2]) < m_Extent[2] ) 
    return true;
		
  return false;
}

//
// Check if a sphere overlaps any part of this bounding box
//
bool CBBox::IsSphereInBox( const Vector &InP, float fRadius)
{
  float fDist;
  float fDistSq = 0;
  Vector P = m_M.InvertSimple() * InP;
	
  // Add distance squared from sphere centerpoint to box for each axis
  for ( int i = 0; i < 3; i++ )
    {
      if ( fabs(P[i]) > m_Extent[i] )
	{
	  fDist = fabs(P[i]) - m_Extent[i];
	  fDistSq += fDist*fDist;
	}
    }
  return ( fDistSq <= fRadius*fRadius );
}

//
// Check if the bounding box is completely behind a plane( defined by a normal and a point )
//
bool CBBox::BoxOutsidePlane( const Vector &InNorm, const Vector &InP )
{
  // Plane Normal in Box Space
  Vector Norm = InNorm.RotByMatrix( m_M.InvertSimple().mf ); // RotByMatrix only uses rotation portion of matrix
  Norm = Vector( fabs( Norm[0] ), fabs( Norm[1] ), fabs( Norm[2] ) );
	
  float Extent = Norm * m_Extent ; // Box Extent along the plane normal
  float Distance = InNorm * ( GetCenterPoint() - InP ); // Distance from Box Center to the Plane

  // If Box Centerpoint is behind the plane further than its extent, the Box is outside the plane
  if ( Distance < -Extent ) return true;
  return false;
}

//
// Does the Line (L1, L2) intersect the Box?
//
bool CBBox::IsLineInBox( const Vector& L1, const Vector& L2 )
{	
  // Put line in box space
  CMatrix MInv = m_M.InvertSimple();
  Vector LB1 = MInv * L1;
  Vector LB2 = MInv * L2;

  // Get line midpoint and extent
  Vector LMid = (LB1 + LB2) * 0.5f; 
  Vector L = (LB1 - LMid);
  Vector LExt = Vector( fabs(L[0]), fabs(L[1]), fabs(L[2]) );

  // Use Separating Axis Test
  // Separation vector from box center to line center is LMid, since the line is in box space
  if ( fabs( LMid[0] ) > m_Extent[0] + LExt[0] ) return false;
  if ( fabs( LMid[1] ) > m_Extent[1] + LExt[1] ) return false;
  if ( fabs( LMid[2] ) > m_Extent[2] + LExt[2] ) return false;
  // Crossproducts of line and each axis
  if ( fabs( LMid[1] * L[2] - LMid[2] * L[1])  >  (m_Extent[1] * LExt[2] + m_Extent[2] * LExt[1]) ) return false;
  if ( fabs( LMid[0] * L[2] - LMid[2] * L[0])  >  (m_Extent[0] * LExt[2] + m_Extent[2] * LExt[0]) ) return false;
  if ( fabs( LMid[0] * L[1] - LMid[1] * L[0])  >  (m_Extent[0] * LExt[1] + m_Extent[1] * LExt[0]) ) return false;
  // No separating axis, the line intersects
  return true;
}

//
// Returns a 3x3 rotation matrix as vectors
//
inline void CBBox::GetInvRot( Vector *pvRot )
{
  pvRot[0] = Vector( m_M.mf[0], m_M.mf[1], m_M.mf[2] );
  pvRot[1] = Vector( m_M.mf[4], m_M.mf[5], m_M.mf[6] );
  pvRot[2] = Vector( m_M.mf[8], m_M.mf[9], m_M.mf[10] );
}

//
// Check if any part of a box is inside any part of another box
// Uses the separating axis test.
//
bool CBBox::IsBoxInBox( CBBox &BBox )
{
  Vector SizeA = m_Extent;
  Vector SizeB = BBox.m_Extent;
  Vector RotA[3], RotB[3];	
  GetInvRot( RotA );
  BBox.GetInvRot( RotB );

  float R[3][3];  // Rotation from B to A
  float AR[3][3]; // absolute values of R matrix, to use with box extents
  float ExtentA, ExtentB, Separation;
  int i, k;

  // Calculate B to A rotation matrix
  for( i = 0; i < 3; i++ )
    for( k = 0; k < 3; k++ )
      {
	R[i][k] = RotA[i]*RotB[k]; 
	AR[i][k] = fabs(R[i][k]);
      }
            
  // Vector separating the centers of Box B and of Box A	
  Vector vSepWS = BBox.GetCenterPoint() - GetCenterPoint();
  // Rotated into Box A's coordinates
  Vector vSepA( vSepWS * RotA[0], vSepWS * RotA[1], vSepWS * RotA[2] );            

  // Test if any of A's basis vectors separate the box
  for( i = 0; i < 3; i++ )
    {
      ExtentA = SizeA[i];
      ExtentB = SizeB * ( Vector( AR[i][0], AR[i][1], AR[i][2] ) );
      Separation = fabs( vSepA[i] );

      if( Separation > ExtentA + ExtentB ) return false;
    }

  // Test if any of B's basis vectors separate the box
  for( k = 0; k < 3; k++ )
    {
      ExtentA = SizeA *( Vector( AR[0][k], AR[1][k], AR[2][k] ) );
      ExtentB = SizeB[k];
      Separation = fabs( vSepA * ( Vector(R[0][k],R[1][k],R[2][k]) ) );

      if( Separation > ExtentA + ExtentB ) return false;
    }

  // Now test Cross Products of each basis vector combination ( A[i], B[k] )
  for( i=0 ; i<3 ; i++ )
    for( k=0 ; k<3 ; k++ )
      {
	int i1 = (i+1)%3, i2 = (i+2)%3;
	int k1 = (k+1)%3, k2 = (k+2)%3;
	ExtentA = SizeA[i1] * AR[i2][k]  +  SizeA[i2] * AR[i1][k];
	ExtentB = SizeB[k1] * AR[i][k2]  +  SizeB[k2] * AR[i][k1];
	Separation = fabs( vSepA[i2] * R[i1][k]  -  vSepA[i1] * R[i2][k] );
	if( Separation > ExtentA + ExtentB ) return false;
      }

  // No separating axis found, the boxes overlap	
  return true;
}
